"""Command line menu selector."""

import curses
import re
from .cursor_control import _cursorInput
from .exceptions import *
from . import _validator

__version__ = "0.10"
__author__ = "xyzpw"
__description__ = "A terminal-based interactive menu selector which is controlled via arrow keys."
__license__ = "MIT"

class MenuSelector:
    """Contains items which can be prompted via a menu selector."""
    def __init__(self, items: list | dict, title: str = None, description: str = None):
        self.items = items
        self.title = title
        self.description = description
    def createMenuString(self, num_sep: str = ")", align: bool = False) -> str:
        menu = ""
        if self.title != None: menu += self.title
        menu += "\n\n%s\n\n" % self.description if self.description != None else ("\n\n" if self.title != None else "")
        lengthiestItem = 0
        if align:
            for item in self.items:
                _length = len(str(item))
                if lengthiestItem < _length:
                    lengthiestItem = int(_length)
        makePadding = lambda itemStr: f"{' '*(lengthiestItem - len(itemStr))}" if lengthiestItem > 1 else ""
        for item in self.items:
            _index = list(self.items).index(item)
            menu += "%s%s %s" % (_index, num_sep, item)
            menu += f" - {makePadding(item)}{self.items.get(item)}\n" if isinstance(self.items, dict) else "\n"
        return menu
    def prompt_select(self, num_sep: str = ")", align: bool = False,
            max_items: tuple = (1, 1), clear: bool = False,
            allow_all: bool = False, center: bool = False) -> list[tuple]:
        """Prompts for an integer while displaying the values associated with that index.

        :param num_sep: separator between number and item, e.g. ) for N)
        :param align: aligns items by adding additional spaces between items and their descriptions
        :param max_items: a range of numbers which represent the number of items required to be selected
            :example:
                (1, 3)
                More than 1 option must be selected, but no greater than 3.
        :param clear: clears the screen prior to displaying the menu
        :param allow_all: adds a select all option to the item list
        :param center: positions the menu selector to the center of the terminal"""
        if allow_all:
            max_items = (max_items[0], len(self.items))
        if not _validator.validateMaxItemsRange(max_items):
            raise MenuItemError("allowed items range is invalid")
        menu = self.createMenuString(num_sep, align)
        if allow_all: menu += "%d%s Select All\n" % (len(self.items), num_sep)
        if clear:
            print("\x1b[H\x1b[2J\x1b[3J", end='') # using ansi codes to clear buffer and entirety of terminal screen
        if center: menu = cursor_control._cursorInput.makeStrCentered(menu, True)
        choiceIndexes = re.findall(r"(?P<index>\d+)(?:,\s?|\s|\Z)", input(menu + "> "))
        if allow_all and str(len(self.items)) in choiceIndexes: choiceIndexes = list(range(len(self.items)))
        if not _validator.validateItemSelectionCount(max_items, choiceIndexes):
            raise MenuItemError("number of items selected is out of range")
        return [(i, list(self.items)[int(i)]) for i in choiceIndexes]
    def arrow_select(self, num_sep: str = ")", align: bool = False,
            arrow: str = "=>", max_items: tuple = (1, 1),
            allow_all: bool = False, center: bool = False):
        """Allows the user to select options with the arrow keys.

        :param num_sep: separator between number and item, e.g. ) for N)
        :param align: aligns items by adding additional spaces between items and their descriptions
        :param arrow: the string which represents the arrow that points to the text which will be selected
        :param max_items: a range of numbers which represent the number of items required to be selected
            :Example:
                (1, 3)
                More than 1 option must be selected, but no greater than 3.
        :param allow_all: adds a select all option to the item list
        :param center: positions the menu selector to the center of the terminal"""
        if allow_all:
            max_items = (max_items[0], len(self.items))
        if not _validator.validateMaxItemsRange(max_items):
            raise MenuItemError("allowed items range is invalid")
        menu = self.createMenuString(num_sep, align)
        if max_items[1] > 1 if max_items[1] != None else True:
            selectedIndexes = curses.wrapper(_cursorInput.cursorArrowMultiselectMenu, menu, max_items[1], allow_all, center)
            usrChoices = [(i, list(self.items)[i]) for i in selectedIndexes] if selectedIndexes != None else []
        else:
            selectedIndexes = curses.wrapper(_cursorInput.cursorArrowMenu, menu, arrow, center)
            usrChoices = [(selectedIndexes, list(self.items)[selectedIndexes])] if selectedIndexes != None else []
        if not _validator.validateItemSelectionCount(max_items, usrChoices) and bool(usrChoices):
            raise MenuItemError("number of items selected is out of range")
        return usrChoices
    def highlight_select(self, center: bool = False, disabled_options: list[int] = [], pages: list[list] = []):
        """Highlights the options at the current line index.

        :param center: positions the menu selector to the center of the terminal
        :param disabled_options: list of indexes indicating options that cannot be selected
            Example:
                [0, 1]
                Disables item indexes 0 and 1 from being selected.
        :param pages: a list of arrays which represent a page containing integers pointing to the indexes of items
            Example:
                [[0, 1], [2, 3]]
                Sets item indexes 0 and 1 into page 1 and indexes 2 and 3 into page 2."""
        if not isinstance(pages, list): raise MenuItemError("pages must be an array")
        if set([i for i in range(len(self.items))]) == set(disabled_options):
            raise MenuItemError("at least one item must be selectable")
        menuComponents = self.items, self.title, self.description
        if pages != []:
            itemsWithoutPage = []
            itemsWithPage = []
            for p in pages:
                if not isinstance(p, list):
                    raise MenuItemError("pages must contain only integers within each array, e.g. [[0,1], [2,3]]")
                if p == []:
                    raise MenuItemError("all pages must contain at least 1 item")
                for i in p:
                    if i not in itemsWithPage: itemsWithPage.append(i)
            itemsWithoutPage = [i for i in range(len(self.items)) if i not in itemsWithPage]
            if itemsWithoutPage != []: pages.append([i for i in itemsWithoutPage]) # Creating new page for items not associated with a page
            selectedIndex = curses.wrapper(_cursorInput.highlightMultiPageSelectMenu, menuComponents, center, disabled_options, pages)
        else:
            selectedIndex = curses.wrapper(_cursorInput.highlightSelectMenu, menuComponents, center, disabled_options)
        usrChoice = [(selectedIndex, list(self.items)[selectedIndex])] if selectedIndex != None else []
        return usrChoice
    def modify_select(self, center: bool = False, options: list[tuple] = [], disabled_options: list[int] = []):
        """A menu selector which allows the user to user to use left/right arrow keys to alter through options.

            :param center: positions the menu selector to the center of the terminal
            :param options: an array containing tuples to represent each option for that index
                Example:
                    [(0, "second option to menu index 0", "third option to menu index 0"), (1, "second option for menu index 1")]
            :param disabled_options: a list of indexes indicating options that cannot be selected
                Example:
                    [0, 1]
                    Disables item indexes 0 and 1 from being selected."""
        menuComponents = self.items, self.title, self.description
        usrChoice = curses.wrapper(_cursorInput.editableMenu, menuComponents, center, options, disabled_options)
        return usrChoice



def validateMaxItemsRange(maxItems: tuple[int]):
    least = maxItems[0] if maxItems[0] != None else 0
    greatest = maxItems[1]
    return True if (True if (greatest==None and least >= 0) else False)==True else (True if least <= greatest and least >= 0 else False)==True

def validateItemSelectionCount(maxItems: tuple, selectedItems: int):
    least = maxItems[0] if maxItems[0] != None else 0
    greatest = maxItems[1]
    itemCount = len(selectedItems)
    if itemCount in range(least, greatest + 1) if greatest != None else (True if greatest==None and least >= 0 and itemCount >= least else False):
        return True
    return False



from . import _cursorInput

def centerArrayItems(usrArray: list) -> list:
    lengthiestStr = 0
    for s in usrArray:
        lengthiestStr = len(s) if len(s) > lengthiestStr else lengthiestStr
    spacing = _cursorInput.getCenteredPos()[0] - lengthiestStr//2
    metaStr = [f"{spacing * ' '}{usrArray[i]}" for i in range(len(usrArray))]
    return metaStr

def getItemSpacing(usrArray: list) -> str:
    lengthiestStr = 0
    for s in usrArray:
        lengthiestStr = len(s) if len(s) > lengthiestStr else lengthiestStr
    spacing = _cursorInput.getCenteredPos()[0] - lengthiestStr//2
    return " "*spacing

def getCenteredNewlines(usrArray: list) -> str:
    lengthiestStr = 0
    for s in usrArray:
        lengthiestStr = len(s) if len(s) > lengthiestStr else lengthiestStr
    newlines = "\n" * (_cursorInput.getCenteredPos()[1]//2 - len(usrArray)//2)
    return newlines




import curses
import re
import shutil
from . import _textHandler

KEYS_QUIT = [ord("q"), ord("Q")]
KEYS_SELECT = [curses.KEY_ENTER, 10]

def getMenuComponents(menuString: str):
    menuLines = menuString.splitlines()
    for l in range(len(menuLines)):
        if re.search(r"^0.\s.*?(?=$|\s)", menuLines[l]) != None:
            meta = menuLines[:l] if l != 0 else None
            menuLines = menuLines[l:]
            break
    if meta != None:
        for i in range(meta.count("")):
            meta.pop(meta.index(""))
    return menuLines, meta

def rewriteMenu(menuString: str, menuArrow: str, currentMenuLine: int):
    menuLines = getMenuComponents(menuString)[0]
    menuLines[currentMenuLine] = "%s %s" % (menuArrow, menuLines[currentMenuLine])
    for i in range(len(menuLines)):
        if i != currentMenuLine:
            menuLines[i] = str(len(menuArrow) * " ") + " %s" % menuLines[i]
    return "\n".join(menuLines)

def cursorArrowMenu(stdscr, menuString: str, menuArrow: str, center: bool = False):
    curses.curs_set(0)
    curses.use_default_colors()
    currentLineIndex = 0
    stdscr.refresh()
    while True:
        stdscr.erase()
        stdscr.refresh()
        menuLines, menuMeta = getMenuComponents(menuString)
        rewrittenMenu = rewriteMenu(menuString, menuArrow, currentLineIndex)
        if menuMeta != None:
            if center:
                menuMeta = _textHandler.centerArrayItems(menuMeta)
            stdscr.addstr("\n\n".join(menuMeta) + "\n\n")
        if center: rewrittenMenu = makeStrCentered(rewrittenMenu, True)
        stdscr.addstr(rewrittenMenu, currentLineIndex)
        keyPressed = stdscr.getch()
        if keyPressed == curses.KEY_UP:
            nextLineIndex = currentLineIndex - 1
            currentLineIndex = nextLineIndex if currentLineIndex > 0 else len(menuLines) - 1
        elif keyPressed == curses.KEY_DOWN:
            nextLineIndex = currentLineIndex + 1
            currentLineIndex = nextLineIndex if currentLineIndex < len(menuLines) - 1 else 0
        elif keyPressed in KEYS_SELECT:
            return currentLineIndex
        elif keyPressed in KEYS_QUIT:
            return

def rewriteMultiselectMenu(menuString: str, currentMenuLine: int, selectedItems: list, allowAll: bool = False):
    menuLines = getMenuComponents(menuString)[0]
    for i in range(len(menuLines)):
        selectionStatusCharacter = "\u25cf" if i in selectedItems else "\u25cb"
        if i == len(menuLines) - 1:
            menuLines[i] = "  %s" % menuLines[i] if currentMenuLine != i else "* %s" % menuLines[i]
            continue
        # Index for "Select All" option
        elif i == len(menuLines) - 2 and allowAll:
            menuLines[i] = "  %s" % menuLines[i] if currentMenuLine != i else "* %s" % menuLines[i]
            continue
        if currentMenuLine != i:
            menuLines[i] = "  %s %s" % (selectionStatusCharacter, menuLines[i])
        elif currentMenuLine == i:
            menuLines[i] = "* %s %s" % (selectionStatusCharacter, menuLines[i])
    return "\n".join(menuLines)

def getCenteredPos():
    return int(shutil.get_terminal_size().columns)//2, int(shutil.get_terminal_size().lines)//2

def makeStrCentered(targetStr: str, eachLine: bool = False) -> str:
    pos = getCenteredPos()
    if eachLine:
        targetStrLines = targetStr.splitlines()
        lengthiestString = 0
        for s in targetStrLines:
            if len(s) > lengthiestString:
                lengthiestString = len(s)
        spacing = " " * (pos[0] - lengthiestString//2)
        centeredStr = "\n" * (pos[1]//2 - len(targetStrLines)//2)
        for l in targetStrLines:
            centeredStr += f"{spacing}{l}\n"
        # centeredStr += "\n" * (pos[1]//2 + len(targetStrLines)//2) # Adds additional spacing (arbitrary)
    else:
        centeredStr = f"{chr(10)*pos[1]}{' '*pos[0]}{targetStr}"
    return centeredStr

def cursorArrowMultiselectMenu(stdscr, menuString: str, maxItemCount: int = None, allowAll: bool = False, center: bool = False):
    curses.curs_set(0)
    curses.use_default_colors()
    currentLineIndex = 0
    selectedItems = []
    if allowAll:
        menuString += "Select All\n"
    menuString += "Confirm Selection\n"
    stdscr.refresh()
    while True:
        stdscr.erase()
        stdscr.refresh()
        menuLines, menuMeta = getMenuComponents(menuString)
        rewrittenMenu = rewriteMultiselectMenu(menuString, currentLineIndex, selectedItems, allowAll)
        if menuMeta != None:
            if center:
                menuMeta = _textHandler.centerArrayItems(menuMeta)
            stdscr.addstr("\n\n".join(menuMeta) + "\n\n")
        if center: rewrittenMenu = makeStrCentered(rewrittenMenu, True)
        stdscr.addstr(rewrittenMenu, currentLineIndex)
        keyPressed = stdscr.getch()
        if keyPressed == curses.KEY_UP:
            nextLineIndex = currentLineIndex - 1
            currentLineIndex = int(nextLineIndex) if currentLineIndex > 0 else len(menuLines) - 1
        elif keyPressed == curses.KEY_DOWN:
            nextLineIndex = currentLineIndex + 1
            currentLineIndex = int(nextLineIndex) if currentLineIndex < len(menuLines) - 1 else 0
        elif keyPressed in KEYS_SELECT:
            if currentLineIndex == len(menuLines) - 2 and allowAll:
                selectedItems = list(range(len(menuLines) - 2)) if selectedItems != list(range(len(menuLines) - 2)) else []
                continue
            if currentLineIndex == len(menuLines) - 1:
                return selectedItems if bool(selectedItems) else None
            else:
                if currentLineIndex not in selectedItems and (maxItemCount <= len(selectedItems) if maxItemCount != None else False):
                    continue
                selectedItems.append(currentLineIndex) if not currentLineIndex in selectedItems else selectedItems.pop(selectedItems.index(currentLineIndex))
        elif keyPressed in KEYS_QUIT:
            return


def highlightSelectMenu(stdscr, menuComponents: tuple, center: bool = False, disabled_options: list[int] = []):
    curses.curs_set(0)
    curses.use_default_colors()
    menuLines: list = menuComponents[0]
    menuTitle, menuDescription = menuComponents[1], menuComponents[2]
    centerSpacing = _textHandler.getItemSpacing(menuLines)
    enabledMenuLines = [i for i in menuLines if menuLines.index(i) not in disabled_options]
    enabledOptions = [i for i in range(len(menuLines)) if i not in disabled_options]
    currentLineIndex = min(enabledOptions)
    while True:
        stdscr.erase()
        stdscr.refresh()
        if menuTitle != None:
            if center: stdscr.addstr(_textHandler.getItemSpacing([menuTitle]))
            stdscr.addstr(menuTitle + "\n\n")
        if menuDescription != None:
            if center: stdscr.addstr(_textHandler.getItemSpacing([menuDescription]))
            stdscr.addstr(menuDescription + "\n\n")
        if center: stdscr.addstr(_textHandler.getCenteredNewlines(menuLines))
        for l in menuLines:
            _strToAdd = "%s\n" % l
            if center: stdscr.addstr(centerSpacing)
            stdscr.addstr(_strToAdd) if menuLines.index(l) != currentLineIndex else stdscr.addstr(_strToAdd, curses.A_REVERSE)
            del _strToAdd
        keyPressed = stdscr.getch()
        if keyPressed == curses.KEY_DOWN:
            currentLineIndex = min([i for i in enabledOptions if i > currentLineIndex]) if currentLineIndex != max(enabledOptions) else min(enabledOptions)
        elif keyPressed == curses.KEY_UP:
            currentLineIndex = max([i for i in enabledOptions if i < currentLineIndex]) if currentLineIndex != min(enabledOptions) else max(enabledOptions)
        elif keyPressed in KEYS_SELECT:
            return currentLineIndex
        elif keyPressed in KEYS_QUIT:
            return

def highlightMultiPageSelectMenu(stdscr, menuComponents: tuple, center: bool = False,
                                 disabled_options: list[int] = [], pages: list[list] = []):
    curses.curs_set(0)
    curses.use_default_colors()
    menuLines = menuComponents[0]
    menuTitle = menuComponents[1]
    menuDescription = menuComponents[2]
    currentPageNumber, currentPageIndex = 1, 0
    enabledOptions = [i for i in range(len(menuLines)) if i not in disabled_options]
    enabledPageOptions = []
    for p in pages:
        enabledPageOptions.append([i for i in p if i in enabledOptions])
    currentLineIndex = min(enabledOptions)
    while True:
        stdscr.erase()
        stdscr.refresh()
        pageDisplayText = f"Page {currentPageNumber} of {len(pages)}"
        if center: stdscr.addstr(_textHandler.getItemSpacing([pageDisplayText]))
        stdscr.addstr(pageDisplayText + "\n\n")
        if menuTitle != None:
            if center: stdscr.addstr(_textHandler.getItemSpacing([menuTitle]))
            stdscr.addstr(menuTitle + "\n\n")
        if menuDescription != None:
            if center: stdscr.addstr(_textHandler.getItemSpacing([menuDescription]))
            stdscr.addstr(menuDescription + "\n\n")
        for l in pages[currentPageIndex]:
            _strToAdd = "%s\n" % menuLines[l]
            if center: stdscr.addstr(_textHandler.getItemSpacing(_strToAdd))
            stdscr.addstr(_strToAdd) if currentLineIndex != l else stdscr.addstr(_strToAdd, curses.A_REVERSE)
        pressedKey = stdscr.getch()
        workingEnabledPageOptions = enabledPageOptions[currentPageIndex]
        if pressedKey == curses.KEY_DOWN:
            currentLineIndex = min([i for i in workingEnabledPageOptions if i > currentLineIndex]) if currentLineIndex != max(workingEnabledPageOptions) else min(workingEnabledPageOptions)
        elif pressedKey == curses.KEY_UP:
            currentLineIndex = max([i for i in range(min(workingEnabledPageOptions), currentLineIndex)]) if currentLineIndex != min(workingEnabledPageOptions) else max(workingEnabledPageOptions)
        elif pressedKey == curses.KEY_RIGHT:
            currentPageIndex = (currentPageIndex + 1) if currentPageIndex != len(pages)-1 else 0
            currentPageNumber = currentPageIndex + 1
            currentLineIndex = min(enabledPageOptions[currentPageIndex])
        elif pressedKey == curses.KEY_LEFT:
            currentPageIndex = (currentPageIndex - 1) if currentPageIndex != 0 else len(pages)-1
            currentPageNumber = currentPageIndex + 1
            currentLineIndex = min(enabledPageOptions[currentPageIndex])
        elif pressedKey in KEYS_SELECT:
            return currentLineIndex
        elif pressedKey in KEYS_QUIT:
            return

def editableMenu(stdscr, menuComponents: dict, center: bool = False, options: list[tuple] = [], disabled_options: list[int] = []):
    menuItems, menuTitle, menuDescription = menuComponents[0], menuComponents[1], menuComponents[2]
    enabledIndexes = [i for i in range(len(menuItems)) if not i in disabled_options]
    currentLineIndex = min(enabledIndexes)
    checkItemModifiable = lambda _itemIndex: _itemIndex in [i[0] for i in options]
    getItemOptionCount = lambda _itemIndex: len([i for i in options if i[0] == _itemIndex][0]) # don't subtract 1 to account for default value
    modifiableItems = {}
    for i in menuItems:
        if checkItemModifiable(menuItems.index(i)):
            # Creating a dictionary — named `modifiableItems` — containing info for each modifiable item:
                # key is the index out of all items in the menu selector
                # key value is the index to the options value tuple value which contains the menu item string
                # option list example:
                    # [(0, "second selectable item text"), (1, "second selectable item text for second menu item")]
            modifiableItems[menuItems.index(i)] = 0
    curses.curs_set(0)
    curses.use_default_colors()
    while True:
        stdscr.refresh()
        stdscr.erase()
        if menuTitle != None:
            if center: stdscr.addstr(_textHandler.getItemSpacing([menuTitle]))
            stdscr.addstr("%s\n\n" % menuTitle)
        if menuDescription != None:
            if center: stdscr.addstr(_textHandler.getItemSpacing([menuDescription]))
            stdscr.addstr("%s\n\n" % menuDescription)
        for i in menuItems:
            if checkItemModifiable(menuItems.index(i)):
                _currentModifiableStringIndex = modifiableItems[menuItems.index(i)]
                #NOTE: displayed option is default if the index is 0
                if _currentModifiableStringIndex != 0:
                    _currentModifiableString = [o[_currentModifiableStringIndex] for o in options if o[0] == menuItems.index(i)][0]
                else:
                    _currentModifiableString = str(i)
                _strToAdd = "< %s >\n" % _currentModifiableString
            else:
                _strToAdd = "%s\n" % i
            if center: stdscr.addstr(_textHandler.getItemSpacing([i]))
            stdscr.addstr(_strToAdd) if currentLineIndex != menuItems.index(i) else stdscr.addstr(_strToAdd, curses.A_REVERSE)
        pressedKey = stdscr.getch()
        if pressedKey == curses.KEY_DOWN:
            currentLineIndex = min([i for i in range(currentLineIndex+1, len(menuItems)) if not i in disabled_options]) if currentLineIndex != max(enabledIndexes) else min(enabledIndexes)
        elif pressedKey == curses.KEY_UP:
            currentLineIndex = max([i for i in range(currentLineIndex) if not i in disabled_options]) if currentLineIndex != min(enabledIndexes) else max(enabledIndexes)
        elif pressedKey == curses.KEY_RIGHT:
            if not checkItemModifiable(currentLineIndex): continue
            if modifiableItems[currentLineIndex] + 1 > getItemOptionCount(currentLineIndex) - 1:
                modifiableItems[currentLineIndex] = 0
                continue
            modifiableItems[currentLineIndex] += 1
        elif pressedKey == curses.KEY_LEFT:
            if not checkItemModifiable(currentLineIndex): continue
            if modifiableItems[currentLineIndex] - 1 < 0:
                modifiableItems[currentLineIndex] = getItemOptionCount(currentLineIndex) - 1
                continue
            modifiableItems[currentLineIndex] -= 1
        elif pressedKey in KEYS_SELECT:
            if checkItemModifiable(currentLineIndex):
                _currentModifiableString = [i[modifiableItems[currentLineIndex]] for i in options if i[0] == currentLineIndex][0]
                if modifiableItems[currentLineIndex] == 0: _currentModifiableString = menuItems[currentLineIndex]
                return [(currentLineIndex, _currentModifiableString)]
            else:
                return [(currentLineIndex, menuItems[currentLineIndex])]
        elif pressedKey in KEYS_QUIT:
            return



from setuptools import setup
import menuchoice

with open("README.md", "r") as f:
    readme = f.read()

setup(
    name="menuchoice",
    version=menuchoice.__version__,
    description=menuchoice.__description__,
    long_description=readme,
    long_description_content_type="text/markdown",
    author=menuchoice.__author__,
    maintainer=menuchoice.__author__,
    url="https://github.com/xyzpw/menuchoice/",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Environment :: Console",
        "Environment :: Console :: Curses",
        "Intended Audience :: Developers",
    ],
    keywords=[
        "menu selector",
        "item selector",
        "ansi",
    ],
    license=menuchoice.__license__,
)
